@@ -9,7 +9,7 @@ Kafka - Apache Kafka interface for Perl.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka> package version 1.001012 .
+This documentation refers to C<Kafka> package version 1.001013 .
 
 =cut
 
@@ -21,7 +21,7 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 use Exporter qw(
     import
@@ -61,6 +61,7 @@ our @EXPORT_OK = qw(
     $ERROR_METADATA_ATTRIBUTES
     $ERROR_MISMATCH_ARGUMENT
     $ERROR_MISMATCH_CORRELATIONID
+    $ERROR_NO_CONNECTION
     $ERROR_NO_ERROR
     $ERROR_NO_KNOWN_BROKERS
     $ERROR_NOT_BINARY_STRING
@@ -127,12 +128,9 @@ our @EXPORT_OK = qw(
     $NOT_SEND_ANY_RESPONSE
     $RECEIVE_EARLIEST_OFFSETS
     $RECEIVE_LATEST_OFFSET
-    $RECEIVE_MAX_ATTEMPTS
-    $RECEIVE_MAX_RETRIES
     $REQUEST_TIMEOUT
     $RETRY_BACKOFF
     $SEND_MAX_ATTEMPTS
-    $SEND_MAX_RETRIES
     $WAIT_WRITTEN_TO_LOCAL_LOG
 );
 
@@ -570,18 +568,8 @@ This property specifies the number of attempts to send of a message.
 Do not use C<$Kafka::SEND_MAX_ATTEMPTS> in C<Kafka::Producer-<gt>send> request to prevent duplicates.
 
 =cut
-const our $SEND_MAX_RETRIES                     => 4;   # legacy, will be removed in future releases
 const our $SEND_MAX_ATTEMPTS                    => 4;
 
-=item C<$RECEIVE_MAX_ATTEMPTS>
-
-4 - The leader may be unavailable transiently, which can fail the receiving of a response.
-This property specifies the number of attempts to receive of a response.
-
-=cut
-const our $RECEIVE_MAX_RETRIES                  => 4;   # legacy, will be removed in future releases
-const our $RECEIVE_MAX_ATTEMPTS                 => 4;
-
 =item C<$RETRY_BACKOFF>
 
 200 - (ms)
@@ -671,12 +659,13 @@ const our $BLOCK_UNTIL_IS_COMMITTED              => -1;
 
 =item C<$DEFAULT_MAX_WAIT_TIME>
 
-The maximum amount of time (ms) to wait when no sufficient amount of data is available at the time the request is dispatched.
+The maximum amount of time (seconds, may be fractional) to wait when no sufficient amount of data is available
+at the time the request is dispatched.
 
-100 - the server will block until the message is committed by all in sync replicas before sending a response.
+0.1 - allow the server to wait up to 0.1s to try to accumulate data before responding.
 
 =cut
-const our $DEFAULT_MAX_WAIT_TIME                => 100;
+const our $DEFAULT_MAX_WAIT_TIME                => 0.1;
 
 =item C<$MESSAGE_SIZE_OVERHEAD>
 
@@ -778,7 +767,7 @@ const our $COMPRESSION_SNAPPY           => 2;
 
 =head2 Error codes
 
-Possible error codes (complies with a hash of descriptions C<$ERROR>):
+Possible error codes (corresponds to descriptions in C<%ERROR>):
 
 =over
 
@@ -791,7 +780,7 @@ const our $ERROR_MISMATCH_ARGUMENT              => -1000;
 
 =item C<$ERROR_CANNOT_SEND>
 
--1001 - Can't send
+-1001 - Cannot send
 
 =cut
 const our $ERROR_CANNOT_SEND                    => -1001;
@@ -805,14 +794,14 @@ const our $ERROR_SEND_NO_ACK                    => -1002;
 
 =item C<ERROR_CANNOT_RECV>
 
--1003 - Can't recv
+-1003 - Cannot receive
 
 =cut
 const our $ERROR_CANNOT_RECV                    => -1003;
 
 =item C<ERROR_CANNOT_BIND>
 
--1004 - Can't bind
+-1004 - Cannot bind
 
 =cut
 const our $ERROR_CANNOT_BIND                    => -1004;
@@ -833,7 +822,7 @@ const our $ERROR_UNKNOWN_APIKEY                 => -1006;
 
 =item C<$ERROR_CANNOT_GET_METADATA>
 
--1007 - Can't get Metadata
+-1007 - Cannot get Metadata
 
 =cut
 const our $ERROR_CANNOT_GET_METADATA            => -1007;
@@ -894,7 +883,7 @@ const our $ERROR_NOT_BINARY_STRING              => -1014;
 =cut
 const our $ERROR_COMPRESSION                    => -1015;
 
-=item C<$ERROR_RESPOSE_BODY_NOT_RECEIVED>
+=item C<$ERROR_RESPONSEMESSAGE_NOT_RECEIVED>
 
 -1016 - 'ResponseMessage' not received
 
@@ -908,6 +897,14 @@ const our $ERROR_RESPONSEMESSAGE_NOT_RECEIVED   => -1016;
 =cut
 const our $ERROR_INCOMPATIBLE_HOST_IP_VERSION   => -1017;
 
+
+=item C<$ERROR_NO_CONNECTION>
+
+-1018 - No IO connection
+
+=cut
+const our $ERROR_NO_CONNECTION   => -1018;
+
 =back
 
 Contains the descriptions of possible error codes obtained via ERROR_CODE box
@@ -1286,13 +1283,13 @@ Contains the descriptions for possible error codes.
 our %ERROR = (
     # Errors fixed by Kafka package
     $ERROR_MISMATCH_ARGUMENT                        => q{Invalid argument},
-    $ERROR_CANNOT_SEND                              => q{Can't send},
+    $ERROR_CANNOT_SEND                              => q{Cannot send},
     $ERROR_SEND_NO_ACK                              => q{No acknowledgement for sent request},
-    $ERROR_CANNOT_RECV                              => q{Can't recv},
-    $ERROR_CANNOT_BIND                              => q{Can't bind},
+    $ERROR_CANNOT_RECV                              => q{Cannot receive},
+    $ERROR_CANNOT_BIND                              => q{Cannot bind},
     $ERROR_METADATA_ATTRIBUTES                      => q{Unknown metadata attributes},
     $ERROR_UNKNOWN_APIKEY                           => q{Unknown ApiKey},
-    $ERROR_CANNOT_GET_METADATA                      => q{Can't get metadata},
+    $ERROR_CANNOT_GET_METADATA                      => q{Cannot get metadata},
     $ERROR_LEADER_NOT_FOUND                         => q{Leader not found},
     $ERROR_MISMATCH_CORRELATIONID                   => q{Mismatch CorrelationId},
     $ERROR_NO_KNOWN_BROKERS                         => q{There are no known brokers},
@@ -1303,6 +1300,7 @@ our %ERROR = (
     $ERROR_COMPRESSION                              => q{Compression error},
     $ERROR_RESPONSEMESSAGE_NOT_RECEIVED             => q{'ResponseMessage' not received},
     $ERROR_INCOMPATIBLE_HOST_IP_VERSION             => q{'Incompatible host name and IP version'},
+    $ERROR_NO_CONNECTION                            => q{'No connection'},
 
     #-- The Protocol Error Messages
     # https://cwiki.apache.org/confluence/display/KAFKA/A+Guide+To+The+Kafka+Protocol#AGuideToTheKafkaProtocol-ErrorCodes
@@ -1611,6 +1609,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
