@@ -6,7 +6,7 @@ Kafka::Connection - Object interface to connect to a kafka cluster.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::Connection> version 1.001012 .
+This documentation refers to C<Kafka::Connection> version 1.001013 .
 
 =cut
 
@@ -20,7 +20,7 @@ use warnings;
 
 our $DEBUG = 0;
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 use Exporter qw(
     import
@@ -31,8 +31,6 @@ our @EXPORT = qw(
 
 #-- load the modules -----------------------------------------------------------
 
-use Carp;
-use Data::Dumper ();
 use Data::Validate::Domain qw(
     is_hostname
 );
@@ -62,9 +60,7 @@ use Scalar::Util::Numeric qw(
 use Storable qw(
     dclone
 );
-use Time::HiRes qw(
-    sleep
-);
+use Time::HiRes ();
 use Try::Tiny;
 
 use Kafka qw(
@@ -103,12 +99,13 @@ use Kafka qw(
     $ERROR_RESPONSEMESSAGE_NOT_RECEIVED
     $ERROR_SEND_NO_ACK
     $ERROR_UNKNOWN_APIKEY
+    $ERROR_INCOMPATIBLE_HOST_IP_VERSION
+    $ERROR_NO_CONNECTION
 
     $IP_V4
     $IP_V6
     $KAFKA_SERVER_PORT
     $NOT_SEND_ANY_RESPONSE
-    $RECEIVE_MAX_ATTEMPTS
     $REQUEST_TIMEOUT
     $RETRY_BACKOFF
     $SEND_MAX_ATTEMPTS
@@ -266,6 +263,7 @@ const our %RETRY_ON_ERRORS => (
 #    $ERROR_UNSUPPORTED_SASL_MECHANISM       => 1,   # 33 - The broker does not support the requested SASL mechanism
 #    $ERROR_ILLEGAL_SASL_STATE               => 1,   # 34 - Request is not valid given the current SASL state
 #    $ERROR_UNSUPPORTED_VERSION              => 1,   # 35 - The version of API is not supported
+     $ERROR_NO_CONNECTION                    => 1,   # may be disconnected due to idle timeout etc.
 );
 
 #-- constructor ----------------------------------------------------------------
@@ -346,18 +344,6 @@ Default C<$REQUEST_TIMEOUT> is used for the rest of IO operations.
 
 =over 3
 
-=item C<CorrelationId =E<gt> $correlation_id>
-
-Optional, default = C<undef> .
-
-C<Correlation> is a user-supplied integer. It will be passed back with the response by
-the server, unmodified. The C<$correlation_id> should be an integer number.
-
-An exception is thrown if C<CorrelationId> in response does not match the one supplied
-in request.
-
-If C<CorrelationId> is not provided, it is set to a random negative integer.
-
 =item C<SEND_MAX_ATTEMPTS =E<gt> $attempts>
 
 Optional, int32 signed integer, default = C<$Kafka::SEND_MAX_ATTEMPTS> .
@@ -366,17 +352,6 @@ In some circumstances (leader is temporarily unavailable, outdated metadata, etc
 This property specifies the maximum number of attempts to send a message.
 The C<$attempts> should be an integer number.
 
-Do not use C<$Kafka::SEND_MAX_ATTEMPTS> in C<Kafka::Producer-<gt>send> request to prevent duplicates.
-
-=item C<RECEIVE_MAX_ATTEMPTS =E<gt> $attempts>
-
-Optional, int32 signed integer, default = C<$Kafka::RECEIVE_MAX_ATTEMPTS> .
-
-In some circumstances (temporarily network issues, server high load, socket error, etc) we may fail to
-receive a response.
-This property specifies the maximum number of attempts to receive a message.
-The C<$attempts> should be an integer number.
-
 =item C<RETRY_BACKOFF =E<gt> $backoff>
 
 Optional, default = C<$Kafka::RETRY_BACKOFF> .
@@ -422,9 +397,7 @@ sub new {
         broker_list             => [],
         timeout                 => $REQUEST_TIMEOUT,
         ip_version              => undef,
-        CorrelationId           => undef,
         SEND_MAX_ATTEMPTS       => $SEND_MAX_ATTEMPTS,
-        RECEIVE_MAX_ATTEMPTS    => $RECEIVE_MAX_ATTEMPTS,
         RETRY_BACKOFF           => $RETRY_BACKOFF,
         AutoCreateTopicsEnable  => 0,
         MaxLoggedErrors         => 100,
@@ -432,35 +405,19 @@ sub new {
 
     while ( @args ) {
         my $k = shift @args;
-
-        # legacy, will be removed in future releases
-        if ( $k eq 'SEND_MAX_RETRIES' ) {
-            carp "Parameter 'SEND_MAX_RETRIES' is deprecated, use 'SEND_MAX_ATTEMPTS' instead";
-            $k = 'SEND_MAX_ATTEMPTS';
-        } elsif ( $k eq 'RECEIVE_MAX_RETRIES' ) {
-            $k = 'RECEIVE_MAX_ATTEMPTS';
-            carp "Parameter 'RECEIVE_MAX_RETRIES' is deprecated, use 'RECEIVE_MAX_ATTEMPTS' instead";
-        }
-
         $self->{ $k } = shift @args if exists $self->{ $k };
     }
 
-    $self->{CorrelationId} //= _get_CorrelationId;
-
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'host' )
         unless defined( $self->{host} ) && ( $self->{host} eq q{} || defined( _STRING( $self->{host} ) ) ) && !utf8::is_utf8( $self->{host} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'port' )
         unless _POSINT( $self->{port} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'timeout (%s)', $self->{timeout} ) )
-        unless ( _NUMBER( $self->{timeout} ) && $self->{timeout} > 0 && $self->{timeout} <= $MAX_INT32 ) || !defined( $self->{timeout} );
+        unless !defined( $self->{timeout} ) || ( defined _NUMBER( $self->{timeout} ) && int( 1000 * $self->{timeout} ) >= 1 && int( $self->{timeout} * 1000 ) <= $MAX_INT32 );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'broker_list' )
         unless _ARRAY0( $self->{broker_list} );
-    $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'CorrelationId (%s)', $self->{CorrelationId} ) )
-        unless isint( $self->{CorrelationId} ) && $self->{CorrelationId} <= $MAX_CORRELATIONID;
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'SEND_MAX_ATTEMPTS' )
         unless _POSINT( $self->{SEND_MAX_ATTEMPTS} );
-    $self->_error( $ERROR_MISMATCH_ARGUMENT, 'RECEIVE_MAX_ATTEMPTS' )
-        unless _POSINT( $self->{RECEIVE_MAX_ATTEMPTS} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'RETRY_BACKOFF' )
         unless _POSINT( $self->{RETRY_BACKOFF} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'MaxLoggedErrors' )
@@ -577,7 +534,7 @@ sub get_metadata {
 
     $self->_update_metadata( $topic )
         # FATAL error
-        or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic = '%s'", $topic ) );
+        or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic='%s'", $topic ) );
 
     my $clone;
     if ( defined $topic ) {
@@ -605,13 +562,10 @@ sub is_server_known {
     return exists $self->{_IO_cache}->{ $server };
 }
 
-=head3 C<is_server_alive( $server )>
-
-Returns true, if known C<$server> (host:port or [IPv6_host]:port) is accessible.
-Checks the accessibility of the server.
-
-=cut
-sub is_server_alive {
+# Returns true, if known C<$server> (host:port or [IPv6_host]:port) is accessible.
+# Checks the accessibility of the server.
+# This is evil: opens and closes a NEW connection immediately so do not use unless there is a strong reason for it.
+sub _is_server_alive {
     my ( $self, $server ) = @_;
 
     $self->_error( $ERROR_MISMATCH_ARGUMENT )
@@ -625,18 +579,14 @@ sub is_server_alive {
         unless exists( $io_cache->{ $server } );
 
     if ( my $io = $self->_connectIO( $server ) ) {
-        return $io->is_alive;
+        return $io->_is_alive;
     } else {
         return;
     }
 }
 
-=head3 C<is_server_connected( $server )>
-
-Returns true, if successful connection is established with C<$server> (host:port or [IPv6_host]:port).
-
-=cut
-sub is_server_connected {
+# this is evil, do not use unless there is a very strong reason for it
+sub _is_server_connected {
     my ( $self, $server ) = @_;
 
     $self->_error( $ERROR_MISMATCH_ARGUMENT )
@@ -648,7 +598,7 @@ sub is_server_connected {
         return;
     }
 
-    return $io->is_alive;
+    return $io->_is_alive;
 }
 
 =head3 C<receive_response_to_request( $request, $compression_codec )>
@@ -697,14 +647,12 @@ L<$COMPRESSION_SNAPPY|Kafka/$COMPRESSION_SNAPPY>.
 
 =cut
 sub receive_response_to_request {
-    my ( $self, $request, $compression_codec ) = @_;
-
-    local $Data::Dumper::Sortkeys = 1 if $self->debug_level;
+    my ( $self, $request, $compression_codec, $response_timeout ) = @_;
 
     my $api_key = $request->{ApiKey};
 
-# WARNING: The current version of the module limited to the following:
-# supports queries with only one combination of topic + partition (first and only).
+    # WARNING: The current version of the module limited to the following:
+    # supports queries with only one combination of topic + partition (first and only).
 
     my $topic_data  = $request->{topics}->[0];
     my $topic_name  = $topic_data->{TopicName};
@@ -716,127 +664,154 @@ sub receive_response_to_request {
     ) {
         $self->_update_metadata( $topic_name )  # hash metadata could be updated
             # FATAL error
-            or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic = '%s'", $topic_name ) );
+            or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic='%s'", $topic_name ), request => $request )
+        ;
     }
+
+    $request->{CorrelationId} = _get_CorrelationId() unless exists $request->{CorrelationId};
+
     my $encoded_request = $protocol{ $api_key }->{encode}->( $request, $compression_codec );
 
-    my $CorrelationId = $request->{CorrelationId} // _get_CorrelationId;
+    say STDERR format_message( '[%s] compression_codec=%s, request=%s',
+        scalar( localtime ),
+        $compression_codec,
+        $request,
+    ) if $self->debug_level;
 
-    say STDERR sprintf( '[%s] compression_codec = %d, request: %s',
-            scalar( localtime ),
-            $compression_codec // '<undef>',
-            Data::Dumper->Dump( [ $request ], [ 'request' ] )
-        ) if $self->debug_level;
+    my( $ErrorCode, $partition_data );
 
-    my $attempts = $api_key == $APIKEY_PRODUCE ? 1 : $self->{SEND_MAX_ATTEMPTS};
-    my ( $ErrorCode, $partition_data, $server );
-    ATTEMPTS:
-    while ( $attempts-- ) {
-        REQUEST:
-        {
-            $ErrorCode = $ERROR_NO_ERROR;
-            if ( defined( my $leader = $self->{_metadata}->{ $topic_name }->{ $partition }->{Leader} ) ) {   # hash metadata could be updated
-                unless ( $server = $self->{_leaders}->{ $leader } ) {
-                    $ErrorCode = $ERROR_LEADER_NOT_FOUND;
-                    $self->_remember_nonfatal_error( $ErrorCode, $ERROR{ $ErrorCode }, $server, $topic_name, $partition );
-                    last REQUEST;       # go to the next attempt
-                }
+    my $attempt = 0;
+    ATTEMPT: while ( ++$attempt <= ( $self->{SEND_MAX_ATTEMPTS} // 1 ) ) {
+        $ErrorCode = $ERROR_NO_ERROR;
+        # hash metadata could be updated
+        my $leader = $self->{_metadata}->{ $topic_name }->{ $partition }->{Leader};
+        next ATTEMPT unless defined $leader;
 
-                # Send a request to the leader
-                if ( !$self->_connectIO( $server ) ) {
-                    $ErrorCode = $ERROR_CANNOT_BIND;
-                } elsif ( !$self->_sendIO( $server, $encoded_request ) ) {
-                    $ErrorCode = $ERROR_CANNOT_SEND;
-                }
-                if ( $ErrorCode != $ERROR_NO_ERROR ) {
-                    $self->_remember_nonfatal_error( $ErrorCode, $self->{_IO_cache}->{ $server }->{error}, $server, $topic_name, $partition );
-                    last REQUEST;    # go to the next attempt
-                }
+        my $server = $self->{_leaders}->{ $leader };
+        unless ( $server ) {
+            $ErrorCode = $ERROR_LEADER_NOT_FOUND;
+            $self->_remember_nonfatal_error( $ErrorCode, $ERROR{ $ErrorCode }, $server, $topic_name, $partition );
+            next ATTEMPT;
+        }
 
-                my $response;
-                if ( $api_key == $APIKEY_PRODUCE && $request->{RequiredAcks} == $NOT_SEND_ANY_RESPONSE ) {
+        # Send a request to the leader
+        if ( $self->_connectIO( $server ) ) {
+            unless ( $self->_sendIO( $server, $encoded_request ) ) {
+                $ErrorCode = $self->_io_error_code( $server );
+                $ErrorCode = $ERROR_CANNOT_SEND if $ErrorCode == $ERROR_NO_ERROR;
+                $self->_closeIO( $server, 1 );
+            }
+        }
+        else {
+            $ErrorCode = $ERROR_CANNOT_BIND;
+        }
 
-                    # Do not receive a response, self-forming own response
-                    $response = {
-                        CorrelationId                           => $CorrelationId,
-                        topics                                  => [
+        if ( $ErrorCode != $ERROR_NO_ERROR ) {
+            # could not send request due to non-fatal IO error (fatal errors should be thrown by connectIO/sendIO already)
+            $self->_remember_nonfatal_error( $ErrorCode, $self->_io_error( $server ), $server, $topic_name, $partition );
+            if( $api_key == $APIKEY_PRODUCE && !( $ErrorCode == $ERROR_CANNOT_BIND || $ErrorCode == $ERROR_NO_CONNECTION ) ) {
+                # do not retry failed produce requests which may have sent some data already
+                $ErrorCode = $ERROR_CANNOT_SEND;
+                last ATTEMPT;
+            }
+            next ATTEMPT;
+        }
+
+        my $response;
+        if ( $api_key == $APIKEY_PRODUCE && $request->{RequiredAcks} == $NOT_SEND_ANY_RESPONSE ) {
+            # Do not receive a response, self-forming own response
+            $response = {
+                CorrelationId                           => $request->{CorrelationId},
+                topics                                  => [
+                    {
+                        TopicName                       => $topic_name,
+                        partitions                      => [
                             {
-                                TopicName                       => $topic_name,
-                                partitions                      => [
-                                    {
-                                        Partition               => $partition,
-                                        ErrorCode               => 0,
-                                        Offset                  => $BAD_OFFSET,
-                                    },
-                                ],
+                                Partition               => $partition,
+                                ErrorCode               => 0,
+                                Offset                  => $BAD_OFFSET,
                             },
                         ],
-                    };
-                } else {
-                    my $encoded_response_ref;
-                    unless ( $encoded_response_ref = $self->_receiveIO( $server ) ) {
-                        if ( $api_key == $APIKEY_PRODUCE ) {
-# WARNING: Unfortunately, the sent package (one or more messages) does not have a unique identifier
-# and there is no way to verify the delivery of data
-                            $ErrorCode = $ERROR_SEND_NO_ACK;
+                    },
+                ],
+            };
+        } else {
+            my $encoded_response_ref = $self->_receiveIO( $server, $response_timeout );
+            unless ( $encoded_response_ref ) {
+                if ( $api_key == $APIKEY_PRODUCE ) {
+                    # WARNING: Unfortunately, the sent package (one or more messages) does not have a unique identifier
+                    # and there is no way to verify the delivery of data
+                    $ErrorCode = $ERROR_SEND_NO_ACK;
 
-                            # Should not be allowed to re-send data on the next attempt
-                            # FATAL error
-                            $self->_error( $ErrorCode, $self->{_IO_cache}->{ $server }->{error} );
-                        } else {
-                            $ErrorCode = $ERROR_CANNOT_RECV;
-                            $self->_remember_nonfatal_error( $ErrorCode, $self->{_IO_cache}->{ $server }->{error}, $server, $topic_name, $partition );
-                            last REQUEST;   # go to the next attempt
-                        }
-                    }
-                    if ( length( $$encoded_response_ref ) > 4 ) {   # MessageSize => int32
-                        $response = $protocol{ $api_key }->{decode}->( $encoded_response_ref );
-                        say STDERR sprintf( '[%s] response: %s',
-                                scalar( localtime ),
-                                Data::Dumper->Dump( [ $response ], [ 'response' ] )
-                            ) if $self->debug_level;
-                    } else {
-                        $self->_error( $ERROR_RESPONSEMESSAGE_NOT_RECEIVED );
-                    }
-                }
-
-                $response->{CorrelationId} == $CorrelationId
+                    # Should not be allowed to re-send data on the next attempt
                     # FATAL error
-                    or $self->_error( $ERROR_MISMATCH_CORRELATIONID );
-                $topic_data     = $response->{topics}->[0];
-                $partition_data = $topic_data->{ $api_key == $APIKEY_OFFSET ? 'PartitionOffsets' : 'partitions' }->[0];
-
-                if ( ( $ErrorCode = $partition_data->{ErrorCode} ) == $ERROR_NO_ERROR ) {
-                    return $response;
-                } elsif ( exists $RETRY_ON_ERRORS{ $ErrorCode } ) {
-                    $self->_remember_nonfatal_error( $ErrorCode, $ERROR{ $ErrorCode }, $server, $topic_name, $partition );
-                    last REQUEST;   # go to the next attempt
+                    $self->_error( $ErrorCode, $self->_io_error( $server ), request => $request );
+                    last ATTEMPT;
                 } else {
-                    # FATAL error
-                    $self->_error( $ErrorCode, format_message( "topic = '%s', partition = %s", $topic_name, $partition ) );
+                    $ErrorCode = $ERROR_CANNOT_RECV;
+                    $self->_remember_nonfatal_error( $ErrorCode, $self->_io_error( $server ), $server, $topic_name, $partition );
+                    next ATTEMPT;
                 }
             }
+            if ( length( $$encoded_response_ref ) > 4 ) {   # MessageSize => int32
+                $response = $protocol{ $api_key }->{decode}->( $encoded_response_ref );
+                say STDERR format_message( '[%s] response: %s',
+                        scalar( localtime ),
+                        $response,
+                    ) if $self->debug_level;
+            } else {
+                $self->_error( $ERROR_RESPONSEMESSAGE_NOT_RECEIVED, format_message("response length=%s", length( $$encoded_response_ref ) ), request => $request );
+            }
         }
 
+        # FATAL error if correllation does not match
+        $self->_error( $ERROR_MISMATCH_CORRELATIONID, "$response->{CorrelationId} != $request->{CorrelationId}", request => $request, response => $response )
+            unless $response->{CorrelationId} == $request->{CorrelationId}
+        ;
+        $topic_data     = $response->{topics}->[0];
+        $partition_data = $topic_data->{ $api_key == $APIKEY_OFFSET ? 'PartitionOffsets' : 'partitions' }->[0];
+
+        $ErrorCode = $partition_data->{ErrorCode};
+
+        return $response if $ErrorCode == $ERROR_NO_ERROR; # success
+
+        if( $api_key == $APIKEY_PRODUCE && $ErrorCode == $ERROR_REQUEST_TIMED_OUT ) {
+            # special case: produce request timed out so we did not get expected ACK and should not retry sending request again
+            # Should not be allowed to re-send data on the next attempt
+            # FATAL error
+            $self->_error( $ERROR_SEND_NO_ACK, format_message( "topic='%s', partition=%s response error: %s", $topic_name, $partition, $ErrorCode ), request => $request, response => $response );
+            last ATTEMPT;
+        }
+
+        if ( exists $RETRY_ON_ERRORS{ $ErrorCode } ) {
+            $self->_remember_nonfatal_error( $ErrorCode, $ERROR{ $ErrorCode }, $server, $topic_name, $partition );
+            next ATTEMPT;
+        }
+
+        # FATAL error
+        $self->_error( $ErrorCode, format_message( "topic='%s', partition=%s", $topic_name, $partition ), request => $request );
+    } continue {
         # Expect to possible changes in the situation, such as restoration of connection
-        say STDERR sprintf( '[%s] sleeping for %d ms before making request attempt #%d (%s)',
+        say STDERR format_message( '[%s] sleeping for %d ms before making request attempt #%d (%s)',
                 scalar( localtime ),
                 $self->{RETRY_BACKOFF},
-                $self->{SEND_MAX_ATTEMPTS} - $attempts + 1,
+                $attempt + 1,
                 $ErrorCode == $ERROR_NO_ERROR ? 'refreshing metadata' : "ErrorCode ${ErrorCode}",
             ) if $self->debug_level;
-        sleep $self->{RETRY_BACKOFF} / 1000;
+
+        Time::HiRes::sleep( $self->{RETRY_BACKOFF} / 1000 );
 
         $self->_update_metadata( $topic_name )
             # FATAL error
-            or $self->_error( $ErrorCode || $ERROR_CANNOT_GET_METADATA, format_message( "topic = '%s', partition = %s", $topic_name, $partition ) );
+            or $self->_error( $ErrorCode || $ERROR_CANNOT_GET_METADATA, format_message( "topic='%s', partition=%s", $topic_name, $partition ), request => $request )
+        ;
     }
 
     # FATAL error
     if ( $ErrorCode ) {
-        $self->_error( $ErrorCode, format_message( "topic = '%s'%s", $topic_data->{TopicName}, $partition_data ? ", partition = ".$partition_data->{Partition} : q{} ) );
+        $self->_error( $ErrorCode, format_message( "topic='%s'%s", $topic_data->{TopicName}, $partition_data ? ", partition = ".$partition_data->{Partition} : '' ), request => $request );
     } else {
-        $self->_error( $ERROR_UNKNOWN_TOPIC_OR_PARTITION, format_message( "topic = '%s', partition = %s", $topic_name, $partition ) );
+        $self->_error( $ERROR_UNKNOWN_TOPIC_OR_PARTITION, format_message( "topic='%s', partition=%s", $topic_name, $partition ), request => $request );
     }
 
     return;
@@ -871,7 +846,7 @@ sub exists_topic_partition {
     unless ( %{ $self->{_metadata} } ) {    # the first request
         $self->_update_metadata( $topic )   # hash metadata could be updated
             # FATAL error
-            or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic = '%s'", $topic ) );
+            or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic='%s'", $topic ) );
     }
 
     return exists $self->{_metadata}->{ $topic }->{ $partition };
@@ -922,9 +897,8 @@ sub cluster_errors {
     my ( $self ) = @_;
 
     my %errors;
-    my $io_cache = $self->{_IO_cache};
-    foreach my $server ( keys %$io_cache ) {
-        if ( my $error = $io_cache->{ $server }->{error} ) {
+    foreach my $server ( $self->get_known_servers ) {
+        if ( my $error = $self->_io_error( $server ) ) {
             $errors{ $server } = $error;
         }
     }
@@ -1049,7 +1023,7 @@ sub _get_interviewed_servers {
 sub _update_metadata {
     my ( $self, $topic, $is_recursive_call ) = @_;
 
-    my $CorrelationId = $self->{CorrelationId};
+    my $CorrelationId = _get_CorrelationId();
     my $decoded_request = {
         CorrelationId   => $CorrelationId,
         ClientId        => q{},
@@ -1057,9 +1031,9 @@ sub _update_metadata {
             $topic // (),
         ],
     };
-    say STDERR sprintf( '[%s] metadata request: %s',
+    say STDERR format_message( '[%s] metadata request: %s',
             scalar( localtime ),
-            Data::Dumper->Dump( [ $decoded_request ], [ 'request' ] )
+            $decoded_request,
         ) if $self->debug_level;
     my $encoded_request = $protocol{ $APIKEY_METADATA }->{encode}->( $decoded_request );
 
@@ -1079,9 +1053,9 @@ sub _update_metadata {
     }
 
     my $decoded_response = $protocol{ $APIKEY_METADATA }->{decode}->( $encoded_response_ref );
-    say STDERR sprintf( '[%s] metadata response: %s',
+    say STDERR format_message( '[%s] metadata response: %s',
             scalar( localtime ),
-            Data::Dumper->Dump( [ $decoded_response ], [ 'response' ] )
+            $decoded_response,
         ) if $self->debug_level;
     ( defined( $decoded_response->{CorrelationId} ) && $decoded_response->{CorrelationId} == $CorrelationId )
         # FATAL error
@@ -1092,7 +1066,7 @@ sub _update_metadata {
             return $self->_attempt_update_metadata( $is_recursive_call, $topic, undef, $ERROR_NO_KNOWN_BROKERS );
         } else {
             # FATAL error
-            $self->_error( $ERROR_NO_KNOWN_BROKERS, format_message( "topic = '%s'", $topic ) );
+            $self->_error( $ERROR_NO_KNOWN_BROKERS, format_message( "topic='%s'", $topic ) );
         }
     }
 
@@ -1149,13 +1123,13 @@ sub _update_metadata {
             return $self->_attempt_update_metadata( $is_recursive_call, $TopicName, $partition, $ErrorCode );
         } else {
             # FATAL error
-            $self->_error( $ErrorCode, format_message( "topic = '%s'%s", $TopicName, defined( $partition ) ? ", partition = $partition" : '' ) );
+            $self->_error( $ErrorCode, format_message( "topic='%s'%s", $TopicName, defined( $partition ) ? ", partition=$partition" : '' ) );
         }
     }
 
     %$received_metadata
         # FATAL error
-        or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic = '%s'", $topic ) );
+        or $self->_error( $ERROR_CANNOT_GET_METADATA, format_message( "topic='%s'", $topic ) );
 
     # Update metadata for received topics
     $self->{_metadata}->{ $_ }  = $received_metadata->{ $_ } foreach keys %{ $received_metadata };
@@ -1174,16 +1148,16 @@ sub _attempt_update_metadata {
     my $attempts = $self->{SEND_MAX_ATTEMPTS};
     ATTEMPTS:
     while ( $attempts-- ) {
-        say STDERR sprintf( '[%s] sleeping for %d ms before making update metadata attempt #%d',
+        say STDERR format_message( '[%s] sleeping for %d ms before making update metadata attempt #%d',
                 scalar( localtime ),
                 $self->{RETRY_BACKOFF},
-                $self->{RECEIVE_MAX_ATTEMPTS} - $attempts + 1,
+                $self->{SEND_MAX_ATTEMPTS} - $attempts + 1,
             ) if $self->debug_level;
-        sleep $self->{RETRY_BACKOFF} / 1000;
+        Time::HiRes::sleep( $self->{RETRY_BACKOFF} / 1000 );
         return( 1 ) if $self->_update_metadata( $topic, 1 );
     }
     # FATAL error
-    $self->_error( $error_code, format_message( "topic = '%s'%s", $topic, defined( $partition ) ? ", partition = $partition" : '' ) );
+    $self->_error( $error_code, format_message( "topic='%s'%s", $topic, defined( $partition ) ? ", partition=$partition" : '' ) );
 
     return;
 }
@@ -1200,27 +1174,46 @@ sub _build_server_name {
 # remembers error communicating with the server
 sub _on_io_error {
     my ( $self, $server_data, $error ) = @_;
-
-    my $message;
-    if ( !blessed( $error ) || !$error->isa( 'Kafka::Exception' ) ) {
-        $message = $error;
-    } else {
-        $message = $error->message;
+    $server_data->{error} = $error;
+    if( $server_data->{IO} ) {
+        $server_data->{IO}->close;
+        $server_data->{IO} = undef;
     }
 
-    $server_data->{error}   = $message;
-    $server_data->{IO}      = undef;
+    if( blessed( $error ) && $error->isa('Kafka::Exception') ) {
+        if( $error->code == $ERROR_MISMATCH_ARGUMENT || $error->code == $ERROR_INCOMPATIBLE_HOST_IP_VERSION ) {
+            # rethrow known fatal errors
+            die $error;
+        }
+    } else {
+        # rethrow all unknown errors
+        die $error;
+    }
 
     return;
 }
 
+sub _io_error {
+    my( $self, $server ) = @_;
+    my $server_data = $self->{_IO_cache}->{ $server } or return;
+    return $server_data->{error};
+}
+
+sub _io_error_code {
+    my( $self, $server ) = @_;
+    my $error = $self->_io_error( $server );
+    return $error ? $error->code : $ERROR_NO_ERROR;
+}
+
 # connects to a server (host:port or [IPv6_host]:port)
 sub _connectIO {
     my ( $self, $server ) = @_;
 
-    my $server_data = $self->{_IO_cache}->{ $server };
-    my $io;
-    unless ( $server_data && ( $io = $server_data->{IO} ) ) {
+    my $server_data = $self->{_IO_cache}->{ $server }
+        or $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( "Unknown server '%s' (is not found in the metadata)", $server ) )
+    ;
+    unless( $server_data->{IO} ) {
+        my $error;
         try {
             $server_data->{IO} = Kafka::IO->new(
                 host        => $server_data->{host},
@@ -1230,11 +1223,13 @@ sub _connectIO {
             );
             $server_data->{error} = undef;
         } catch {
-            # NOTE: it is possible to repeat the operation here
-            my $error = $_;
+            $error = $_;
+        };
+
+        if( defined $error ) {
             $self->_on_io_error( $server_data, $error );
             return;
-        };
+        }
     }
 
     return $server_data->{IO};
@@ -1244,63 +1239,66 @@ sub _connectIO {
 sub _sendIO {
     my ( $self, $server, $encoded_request ) = @_;
 
-    my $server_data = $self->{_IO_cache}->{ $server };
+    my $server_data = $self->{_IO_cache}->{ $server }
+        or $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( "Unknown server '%s' (is not found in the metadata)", $server ) )
+    ;
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( "Server '%s' is not connected", $server ) )
+        unless $server_data->{IO}
+    ;
     my $sent;
+    my $error;
     try {
         $sent = $server_data->{IO}->send( $encoded_request );
     } catch {
-        # NOTE: it is possible to repeat the operation here
-        my $error = $_;
-        $self->_on_io_error( $server_data, $error );
+        $error = $_;
     };
 
+    if( defined $error ) {
+        $self->_on_io_error( $server_data, $error );
+    }
+
     return $sent;
 }
 
 # Receive response from a given server
 sub _receiveIO {
-    my ( $self, $server ) = @_;
+    my ( $self, $server, $response_timeout ) = @_;
 
-    my $server_data = $self->{_IO_cache}->{ $server };
+    my $server_data = $self->{_IO_cache}->{ $server }
+        or $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( "Unknown server '%s' (is not found in the metadata)", $server ) )
+    ;
+    my $io = $server_data->{IO}
+        or $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( "Server '%s' is not connected", $server ) )
+    ;
     my $response_ref;
-
     my $error;
-    my $attempts = $self->{RECEIVE_MAX_ATTEMPTS};
-    ATTEMPTS:
-    while ( $attempts-- ) {
-        $error = undef;
-        try {
-            my $io = $server_data->{IO};
-            $response_ref   = $io->receive( 4 ) unless $response_ref;
-            $$response_ref .= ${ $io->receive( unpack( 'l>', $$response_ref ) ) };
-        } catch {
-            $error = $_;
-        };
-        last unless $error;
+    try {
+        $response_ref = $io->receive( 4, $response_timeout ); # response header must arrive within request-specific timeout if provided
+        if ( $response_ref && length( $$response_ref ) == 4 ) {
+            # received 4-byte response header with response size; try receiving the rest
+            my $message_body_ref = $io->receive( unpack( 'l>', $$response_ref ) );
+            $$response_ref .= $$message_body_ref;
+        }
+    } catch {
+        $error = $_;
+    };
 
-        say STDERR sprintf( "[%s] sleeping for %d ms before making receive attempt #%d (error '%s')",
-                scalar( localtime ),
-                $self->{RETRY_BACKOFF},
-                $self->{RECEIVE_MAX_ATTEMPTS} - $attempts + 1,
-                $error,
-            ) if $self->debug_level;
-        sleep $self->{RETRY_BACKOFF} / 1000;
+    if( defined $error ) {
+        $self->_on_io_error( $server_data, $error );
     }
-    $self->_on_io_error( $server_data, $error )
-        if $error;
 
     return $response_ref;
 }
 
 # Close connectino to $server
 sub _closeIO {
-    my ( $self, $server ) = @_;
+    my ( $self, $server, $keep_error ) = @_;
 
     if ( my $server_data = $self->{_IO_cache}->{ $server } ) {
         if ( my $io = $server_data->{IO} ) {
             $io->close;
-            $server_data->{error}   = undef;
-            $server_data->{IO}      = undef;
+            $server_data->{error} = undef unless $keep_error;
+            $server_data->{IO}    = undef;
         }
     }
 
@@ -1330,10 +1328,7 @@ sub _is_like_server {
 # Handler for errors
 sub _error {
     my $self = shift;
-
     Kafka::Exception::Connection->throw( throw_args( @_ ) );
-
-    return;
 }
 
 #-- Closes and cleans up -------------------------------------------------------
@@ -1359,19 +1354,19 @@ Here is the list of possible error messages that C<Kafka::Connection> may produc
 
 Invalid argument was provided to C<new> L<constructor|/CONSTRUCTOR> or to other L<method|/METHODS>.
 
-=item C<Can't send>
+=item C<Cannot send>
 
 Request cannot be sent to Kafka.
 
-=item C<Can't recv>
+=item C<Cannot receive>
 
 Response cannot be received from Kafka.
 
-=item C<Can't bind>
+=item C<Cannot bind>
 
-A successful TCP connection can't be established on given host and port.
+A successful TCP connection cannot be established on given host and port.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 Error detected during parsing of response from Kafka.
 
@@ -1387,7 +1382,7 @@ Mismatch of C<CorrelationId> of request and response.
 
 Failed to locate cluster broker.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 Received meta data is incorrect or missing.
 
@@ -1448,6 +1443,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
