@@ -6,7 +6,7 @@ Kafka::Consumer - Perl interface for Kafka consumer client.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::Consumer> version 1.001012 .
+This documentation refers to C<Kafka::Consumer> version 1.001013 .
 
 =cut
 
@@ -18,7 +18,7 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 #-- load the modules -----------------------------------------------------------
 
@@ -26,13 +26,13 @@ use Carp;
 use Params::Util qw(
     _INSTANCE
     _NONNEGINT
+    _NUMBER
     _POSINT
     _STRING
 );
 use Scalar::Util::Numeric qw(
     isint
 );
-use Try::Tiny;
 
 use Kafka qw(
     $BITS64
@@ -196,19 +196,6 @@ C<new()> takes arguments in key-value pairs. The following arguments are recogni
 C<$connection> is the L<Kafka::Connection|Kafka::Connection> object responsible for communication with
 the Apache Kafka cluster.
 
-=item C<CorrelationId =E<gt> $correlation_id>
-
-Optional, default is C<undef>.
-
-C<Correlation> is a user-supplied integer.
-It will be passed back in the response by the server, unmodified.
-The C<$correlation_id> should be an integer number.
-
-C<CorrelationId> will be auto-assigned (random negative number) if it was not provided on
-creation of L<Kafka::Producer|Kafka::Producer> object.
-
-An exception is thrown if C<CorrelationId> sent with request does not match C<CorrelationId> received in response.
-
 =item C<ClientId =E<gt> $client_id>
 
 This is a user supplied identifier (string) for the client application.
@@ -217,7 +204,7 @@ C<ClientId> will be auto-assigned if not passed in when creating L<Kafka::Produc
 
 =item C<MaxWaitTime =E<gt> $max_time>
 
-The maximum amount of time (ms) to wait when no sufficient data is available at the time the
+The maximum amount of time (seconds, may be fractional) to wait when no sufficient data is available at the time the
 request was issued.
 
 Optional, default is C<$DEFAULT_MAX_WAIT_TIME>.
@@ -225,7 +212,7 @@ Optional, default is C<$DEFAULT_MAX_WAIT_TIME>.
 C<$DEFAULT_MAX_WAIT_TIME> is the default time that can be imported from the
 L<Kafka|Kafka> module.
 
-The C<$max_time> must be a positive integer.
+The C<$max_time> must be a positive number.
 
 =item C<MinBytes =E<gt> $min_bytes>
 
@@ -273,7 +260,6 @@ sub new {
 
     my $self = bless {
         Connection          => undef,
-        CorrelationId       => undef,
         ClientId            => undef,
         MaxWaitTime         => $DEFAULT_MAX_WAIT_TIME,
         MinBytes            => $MIN_BYTES_RESPOND_IMMEDIATELY,
@@ -287,17 +273,14 @@ sub new {
         $self->{ $k } = shift @args if exists $self->{ $k };
     }
 
-    $self->{CorrelationId}  //= _get_CorrelationId;
     $self->{ClientId}       //= 'consumer';
 
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'Connection' )
         unless _INSTANCE( $self->{Connection}, 'Kafka::Connection' );
-    $self->_error( $ERROR_MISMATCH_ARGUMENT, 'CorrelationId' )
-        unless defined( $self->{CorrelationId} ) && isint( $self->{CorrelationId} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'ClientId' )
         unless ( $self->{ClientId} eq q{} || defined( _STRING( $self->{ClientId} ) ) ) && !utf8::is_utf8( $self->{ClientId} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'MaxWaitTime (%s)', $self->{MaxWaitTime} ) )
-        unless defined( $self->{MaxWaitTime} ) && isint( $self->{MaxWaitTime} ) && $self->{MaxWaitTime} > 0 && $self->{MaxWaitTime} <= $MAX_INT32;
+        unless defined( $self->{MaxWaitTime} ) && defined _NUMBER( $self->{MaxWaitTime} ) && int( $self->{MaxWaitTime} * 1000 ) >= 1 && int( $self->{MaxWaitTime} * 1000 ) <= $MAX_INT32;
     $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'MinBytes (%s)', $self->{MinBytes} ) )
         unless ( _isbig( $self->{MinBytes} ) ? ( $self->{MinBytes} >= 0 ) : defined( _NONNEGINT( $self->{MinBytes} ) ) ) && $self->{MinBytes} <= $MAX_INT32;
     $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'MaxBytes (%s)', $self->{MaxBytes} ) )
@@ -369,9 +352,9 @@ sub fetch {
 
     my $request = {
         ApiKey                              => $APIKEY_FETCH,
-        CorrelationId                       => $self->{CorrelationId},
+        CorrelationId                       => _get_CorrelationId(),
         ClientId                            => $self->{ClientId},
-        MaxWaitTime                         => $self->{MaxWaitTime},
+        MaxWaitTime                         => int( $self->{MaxWaitTime} * 1000 ),
         MinBytes                            => $self->{MinBytes},
         topics                              => [
             {
@@ -387,7 +370,7 @@ sub fetch {
         ],
     };
 
-    my $response = $self->{Connection}->receive_response_to_request( $request );
+    my $response = $self->{Connection}->receive_response_to_request( $request, undef, $self->{MaxWaitTime} );
 
     my $messages = [];
     foreach my $received_topic ( @{ $response->{topics} } ) {
@@ -491,7 +474,7 @@ sub offsets {
 
     my $request = {
         ApiKey                              => $APIKEY_OFFSET,
-        CorrelationId                       => $self->{CorrelationId},
+        CorrelationId                       => _get_CorrelationId(),
         ClientId                            => $self->{ClientId},
         topics                              => [
             {
@@ -556,19 +539,19 @@ working with L<Kafka|Kafka> package.
 
 Invalid argument passed to a C<new> L<constructor|/CONSTRUCTOR> or other L<method|/METHODS>.
 
-=item C<Can't send>
+=item C<Cannot send>
 
 Request cannot be sent.
 
-=item C<Can't recv>
+=item C<Cannot receive>
 
 Response cannot be received.
 
-=item C<Can't bind>
+=item C<Cannott bind>
 
 TCP connection can't be established on the given host and port.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 Failed to obtain metadata from Kafka servers.
 
@@ -584,7 +567,7 @@ C<CorrelationId> of response doesn't match one in request.
 
 Resulting metadata has no information about cluster brokers.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 Received metadata has incorrect internal structure.
 
@@ -633,6 +616,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
