@@ -6,7 +6,7 @@ Kafka::Exceptions - Perl Kafka API exception definitions.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::Exceptions> version 1.001012 .
+This documentation refers to C<Kafka::Exceptions> version 1.001013 .
 
 =cut
 
@@ -18,7 +18,9 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.001012';
+our $DEBUG = 0;
+
+our $VERSION = '1.001013';
 
 use Exporter qw(
     import
@@ -35,6 +37,7 @@ use Exception::Class (
     },
     'Kafka::Exception::Connection' => {
         isa     => 'Kafka::Exception',
+        fields  => [ 'request', 'response' ],
     },
     'Kafka::Exception::Consumer' => {
         isa     => 'Kafka::Exception',
@@ -43,6 +46,7 @@ use Exception::Class (
         isa     => 'Kafka::Exception',
     },
     'Kafka::Exception::IO' => {
+        fields  => [ 'errno' ],
         isa     => 'Kafka::Exception',
     },
     'Kafka::Exception::Producer' => {
@@ -60,6 +64,10 @@ use Kafka::Internals qw(
     format_message
 );
 
+if( Kafka::Internals::debug_level( __PACKAGE__ ) ) {
+    Kafka::Exception->Trace(1); # include stack traces
+}
+
 #-- declarations ---------------------------------------------------------------
 
 =head1 SYNOPSIS
@@ -206,11 +214,13 @@ An additional error description that contains information about the encountered
 
 =cut
 sub throw_args {
-    my( $error_code, $description ) = @_;
+    my $error_code = shift;
+    my $description = shift;
 
-    return(
+    return (
         code    => $error_code,
         message => format_message( '%s%s', $ERROR{ $error_code }, $description ? ": $description" : '' ),
+        @_,
     );
 }
 
@@ -267,6 +277,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
