@@ -6,7 +6,7 @@ Kafka::IO - Interface to network communication with the Apache Kafka server.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::IO> version 1.001012 .
+This documentation refers to C<Kafka::IO> version 1.001013 .
 
 =cut
 
@@ -20,12 +20,13 @@ use warnings;
 
 our $DEBUG = 0;
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 #-- load the modules -----------------------------------------------------------
 
 use Carp;
 use Config;
+use Const::Fast;
 use Data::Validate::Domain qw(
     is_hostname
 );
@@ -33,8 +34,18 @@ use Data::Validate::IP qw(
     is_ipv4
     is_ipv6
 );
-use Errno;
+use Errno qw(
+    EAGAIN
+    ECONNRESET
+    EINTR
+    EWOULDBLOCK
+    ETIMEDOUT
+);
 use Fcntl;
+use IO::Select;
+use Params::Util qw(
+    _STRING
+);
 use POSIX qw(
     ceil
 );
@@ -45,6 +56,8 @@ use Socket qw(
     AF_INET
     AF_INET6
     IPPROTO_TCP
+    MSG_DONTWAIT
+    MSG_PEEK
     NI_NUMERICHOST
     NIx_NOSERV
     PF_INET
@@ -65,16 +78,16 @@ use Socket qw(
 use Sys::SigAction qw(
     set_sig_handler
 );
+use Time::HiRes ();
 use Try::Tiny;
 
 use Kafka qw(
-    %ERROR
     $ERROR_CANNOT_BIND
     $ERROR_CANNOT_RECV
     $ERROR_CANNOT_SEND
     $ERROR_MISMATCH_ARGUMENT
-    $ERROR_NOT_BINARY_STRING
     $ERROR_INCOMPATIBLE_HOST_IP_VERSION
+    $ERROR_NO_CONNECTION
     $IP_V4
     $IP_V6
     $KAFKA_SERVER_PORT
@@ -142,6 +155,9 @@ This class allows you to create Kafka 0.9+ clients.
 
 =cut
 
+# Hard limit of IO operation retry attempts, to prevent high CPU usage in IO retry loop
+const my $MAX_RETRIES => 30;
+
 our $_hdr;
 
 #-- constructor ----------------------------------------------------------------
@@ -209,7 +225,7 @@ sub new {
     my ( $class, @args ) = @_;
 
     my $self = bless {
-        host        => q{},
+        host        => '',
         timeout     => $REQUEST_TIMEOUT,
         port        => $KAFKA_SERVER_PORT,
         ip_version  => undef,
@@ -227,13 +243,13 @@ sub new {
     ( $self->{host} ) = $self->{host} =~ /\A(.+)\z/;
     ( $self->{port} ) = $self->{port} =~ /\A(.+)\z/;
 
-    $self->{not_accepted} = 0;
-    $self->{socket} = undef;
+    $self->{socket}     = undef;
+    $self->{_io_select} = undef;
     try {
         $self->_connect();
     } catch {
         my $error = $_;
-        $self->_error( $ERROR_CANNOT_BIND, "->new - $error" );
+        $self->_error( $ERROR_CANNOT_BIND, format_message("Kafka::IO(%s:%s)->new: %s", $self->{host}, $self->{port}, $error ) );
     };
 
     return $self;
@@ -247,71 +263,253 @@ The following methods are provided by C<Kafka::IO> class:
 
 =cut
 
-=head3 C<send( $message )>
+=head3 C<< send( $message <, $timeout> ) >>
 
 Sends a C<$message> to Kafka.
 
 The argument must be a bytes string.
 
+Use optional C<$timeout> argument to override default timeout for this request only.
+
 Returns the number of characters sent.
 
 =cut
 sub send {
-    my ( $self, $message ) = @_;
-
-    ( my $len = length( $message ) ) <= $MAX_SOCKET_REQUEST_BYTES
-        or $self->_error( $ERROR_MISMATCH_ARGUMENT, '->send' );
+    my ( $self, $message, $timeout ) = @_;
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, '->send' )
+        unless defined( _STRING( $message ) )
+    ;
+    my $length = length( $message );
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, '->send' )
+        unless $length <= $MAX_SOCKET_REQUEST_BYTES
+    ;
+    $timeout = $self->{timeout} // $REQUEST_TIMEOUT unless defined $timeout;
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, '->receive' )
+        unless $timeout > 0
+    ;
+    my $select = $self->{_io_select};
+    $self->_error( $ERROR_NO_CONNECTION, 'Attempt to work with a closed socket' ) unless $select;
 
     $self->_debug_msg( $message, 'Request to', 'green' )
-        if $self->debug_level >= 2;
+        if $self->debug_level >= 2
+    ;
+    my $sent = 0;
 
-    # accept not accepted earlier
-    while ( select( my $mask = $self->{_select}, undef, undef, 0 ) ) {
-        my $received = $self->receive( $self->{not_accepted} || 1 );
-        return unless ( $received && defined( $$received ) );
-        $self->{not_accepted} = 0;
+    my $started = Time::HiRes::time();
+    my $until = $started + $timeout;
+
+    my $error_code;
+    my $errno;
+    my $retries = 0;
+    my $interrupts = 0;
+    ATTEMPT: while ( $sent < $length && $retries++ < $MAX_RETRIES ) {
+        my $remaining_time = $until - Time::HiRes::time();
+        last ATTEMPT if $remaining_time <= 0; # timeout expired
+
+        undef $!;
+        my $can_write = $select->can_write( $remaining_time );
+        $errno = $!;
+        if ( $errno ) {
+            if ( $errno == EINTR ) {
+                undef $errno;
+                --$retries; # this attempt does not count
+                ++$interrupts;
+                next ATTEMPT;
+            }
+
+            $self->close;
+
+            last ATTEMPT;
+        }
+
+        if ( $can_write ) {
+            # check for EOF on the first attempt only
+            if ( $retries == 1 && $self->_is_close_wait ) {
+                $self->close;
+                $error_code = $ERROR_NO_CONNECTION;
+                last ATTEMPT;
+            }
+
+            undef $!;
+            my $wrote = CORE::send( $self->{socket}, $message, MSG_DONTWAIT );
+            $errno = $!;
+
+            if( defined $wrote && $wrote > 0 ) {
+                $sent += $wrote;
+                if ( $sent < $length ) {
+                    # remove written data from message
+                    $message = substr( $message, $wrote );
+                }
+            }
+
+            if( $errno ) {
+                if( $errno == EINTR ) {
+                    undef $errno;
+                    --$retries; # this attempt does not count
+                    ++$interrupts;
+                    next ATTEMPT;
+                } elsif (
+                           $errno != EAGAIN
+                        && $errno != EWOULDBLOCK
+                        ## on freebsd, if we got ECONNRESET, it's a timeout from the other side
+                        && !( $errno == ECONNRESET && $^O eq 'freebsd' )
+                    ) {
+                    $self->close;
+                    last ATTEMPT;
+                }
+            }
+
+            last ATTEMPT unless defined $wrote;
+        }
     }
-    $self->{not_accepted} = 0;
 
-    my ( $sent, $mask );
+    unless( !$errno && defined( $sent ) && $sent == $length )
     {
-        last unless ( select( undef, $mask = $self->{_select}, undef, $self->{timeout} // $REQUEST_TIMEOUT ) );
-        $sent += send( $self->{socket}, $message, 0 ) // 0;
-        redo if $sent < $len;
+        $self->_error(
+            $error_code // $ERROR_CANNOT_SEND,
+            format_message( "Kafka::IO(%s)->send: ERRNO=%s ERROR='%s' (length=%s, sent=%s, timeout=%s, retries=%s, interrupts=%s, secs=%.6f)",
+                $self->{host},
+                ( $errno // 0 ) + 0,
+                ( $errno // '<none>' ) . '',
+                $length,
+                $sent,
+                $timeout,
+                $retries,
+                $interrupts,
+                Time::HiRes::time() - $started,
+            )
+        );
     }
 
-    ( defined( $sent ) && $sent == $len )
-        or $self->_error( $ERROR_CANNOT_SEND, "->send - $!" );
-
     return $sent;
 }
 
-=head3 C<receive( $length )>
+=head3 C<< receive( $length <, $timeout> ) >>
 
 Receives a message up to C<$length> size from Kafka.
 
 C<$length> argument must be a positive number.
 
+Use optional C<$timeout> argument to override default timeout for this call only.
+
 Returns a reference to the received message.
 
 =cut
 sub receive {
-    my ( $self, $length ) = @_;
+    my ( $self, $length, $timeout ) = @_;
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, '->receive' )
+        unless $length > 0
+    ;
+    $timeout = $self->{timeout} // $REQUEST_TIMEOUT unless defined $timeout;
+    $self->_error( $ERROR_MISMATCH_ARGUMENT, '->receive' )
+        unless $timeout > 0
+    ;
+    my $select = $self->{_io_select};
+    $self->_error( $ERROR_NO_CONNECTION, 'Attempt to work with a closed socket' ) unless $select;
 
-    my ( $from_recv, $message, $buf, $mask );
-    $message = q{};
-    {
-        last unless ( select( $mask = $self->{_select}, undef, undef, $self->{timeout} // $REQUEST_TIMEOUT ) );
-        $from_recv = recv( $self->{socket}, $buf = q{}, $length, 0 );
-        last if !defined( $from_recv ) || $buf eq q{};
-        $message .= $buf;
-        redo if length( $message ) < $length;
+    my $message = '';
+    my $len_to_read = $length;
+
+    my $started = Time::HiRes::time();
+    my $until = $started + $timeout;
+
+    my $error_code;
+    my $errno;
+    my $retries = 0;
+    my $interrupts = 0;
+    ATTEMPT: while ( $len_to_read > 0 && $retries++ < $MAX_RETRIES ) {
+        my $remaining_time = $until - Time::HiRes::time();
+        last if $remaining_time <= 0; # timeout expired
+
+        undef $!;
+        my $can_read = $select->can_read( $remaining_time );
+        $errno = $!;
+        if ( $errno ) {
+            if ( $errno == EINTR ) {
+                undef $errno;
+                --$retries; # this attempt does not count
+                ++$interrupts;
+                next ATTEMPT;
+            }
+
+            $self->close;
+
+            last ATTEMPT;
+        }
+
+        if ( $can_read ) {
+            my $buf = '';
+            undef $!;
+            my $from_recv = CORE::recv( $self->{socket}, $buf, $len_to_read, MSG_DONTWAIT );
+            $errno = $!;
+
+            if ( defined( $from_recv ) && length( $buf ) ) {
+                $message .= $buf;
+                $len_to_read = $length - length( $message );
+            }
+            if ( $errno ) {
+                if ( $errno == EINTR ) {
+                    undef $errno;
+                    --$retries; # this attempt does not count
+                    ++$interrupts;
+                    next ATTEMPT;
+                } elsif (
+                           $errno != EAGAIN
+                        && $errno != EWOULDBLOCK
+                        ## on freebsd, if we got ECONNRESET, it's a timeout from the other side
+                        && !( $errno == ECONNRESET && $^O eq 'freebsd' )
+                    ) {
+                    $self->close;
+                    last ATTEMPT;
+                }
+            }
+
+            if ( length( $buf ) == 0 ) {
+                if( defined( $from_recv ) && ! $errno ) {
+                    # no error and nothing received with select returning "can read" means EOF: other side closed socket
+                    $self->_debug_msg( 'EOF on receive attempt, closing socket' )
+                        if $self->debug_level;
+                    $self->close;
+
+                    if( length( $message ) == 0 ) {
+                        # we did not receive anything yet, so we may (in some cases) reconnect and try again
+                        $error_code = $ERROR_NO_CONNECTION;
+                    }
+
+                    last ATTEMPT;
+                }
+                # we did not read anything on this attempt: wait a bit before the next one; should not happen, but just in case...
+                if ( my $remaining_attempts = $MAX_RETRIES - $retries ) {
+                    $remaining_time = $until - Time::HiRes::time();
+                    my $micro_seconds = int( $remaining_time * 1e6 / $remaining_attempts );
+                    if ( $micro_seconds > 0 ) {
+                        $micro_seconds = 250_000 if $micro_seconds > 250_000; # prevent long sleeps if total remaining time is big
+                        $self->_debug_msg( format_message( 'sleeping (remaining attempts %d, time %.6f): %d microseconds', $remaining_attempts, $remaining_time, $micro_seconds ) )
+                            if $self->debug_level;
+                        Time::HiRes::usleep( $micro_seconds );
+                    }
+                }
+            }
+        }
     }
-    $self->{not_accepted} = ( $length - length( $message ) ) * ( $self->{not_accepted} >= 0 );
-
-    ( defined( $from_recv ) && !$self->{not_accepted} )
-        or $self->_error( $ERROR_CANNOT_RECV, "->receive - $!" );
 
+    unless( !$errno && length( $message ) >= $length )
+    {
+        $self->_error(
+            $error_code // $ERROR_CANNOT_RECV,
+            format_message( "Kafka::IO(%s)->receive: ERRNO=%s ERROR='%s' (length=%s, received=%s, timeout=%s, retries=%s, interrupts=%s, secs=%.6f)",
+                $self->{host},
+                ( $errno // 0 ) + 0,
+                ( $errno // '<none>' ) . '',
+                $length,
+                length( $message ),
+                $timeout,
+                $retries,
+                $interrupts,
+                Time::HiRes::time() - $started,
+            ),
+        );
+    }
     $self->_debug_msg( $message, 'Response from', 'yellow' )
         if $self->debug_level >= 2;
 
@@ -330,20 +528,30 @@ sub close {
 
     my $ret = 1;
     if ( $self->{socket} ) {
-        $self->{_select} = undef;
         $ret = CORE::close( $self->{socket} );
-        $self->{socket} = undef;
+        $self->{socket}     = undef;
+        $self->{_io_select} = undef;
     }
 
     return $ret;
 }
 
-=head3 C<is_alive>
+sub _is_close_wait {
+    my ( $self ) = @_;
+    return 1 unless $self->{socket} && $self->{_io_select}; # closed already
+    # http://stefan.buettcher.org/cs/conn_closed.html
+    # socket is open; check if we can read, and if we can but recv() cannot peek, it means we got EOF
+    return unless $self->{_io_select}->can_read( 0 ); # we cannot read, but may be able to write
+    my $buf = '';
+    undef $!;
+    my $status = CORE::recv( $self->{socket}, $buf, 1, MSG_DONTWAIT | MSG_PEEK ); # peek, do not remove data from queue
+    # EOF when there is no error, status is defined, but result is empty
+    return ! $! && defined $status && length( $buf ) == 0;
+}
 
-The method verifies whether we are connected to Kafka server.
-
-=cut
-sub is_alive {
+# The method verifies if we can connect to a Kafka broker.
+# This is evil: opens and immediately closes a NEW connection so do not use unless there is a strong reason for it.
+sub _is_alive {
     my ( $self ) = @_;
 
     my $socket = $self->{socket};
@@ -365,7 +573,8 @@ sub is_alive {
 sub _connect {
     my ( $self ) = @_;
 
-    $self->{socket} = undef;
+    $self->{socket}     = undef;
+    $self->{_io_select} = undef;
 
     my $name    = $self->{host};
     my $port    = $self->{port};
@@ -426,12 +635,12 @@ sub _connect {
             }
         } else {
             $ip = $self->_gethostbyname( $name );
-            die( format_message( "not resolve a host name to the IP address: %s\n", $name ) ) unless $ip;
+            die( format_message( "could not resolve host name to IP address: %s\n", $name ) ) unless $ip;
         }
     }
 
     # Create socket.
-    socket( my $connection, $self->{pf}, SOCK_STREAM, getprotobyname( 'tcp' ) ) or die( "socket: $!\n" );
+    socket( my $connection, $self->{pf}, SOCK_STREAM, scalar getprotobyname( 'tcp' ) ) or die( "socket: $!\n" );
 
     # Set autoflushing.
     $_ = select( $connection ); $| = 1; select $_;
@@ -445,31 +654,27 @@ sub _connect {
 
     # Connect returns immediately because of O_NONBLOCK.
     my $sockaddr = $self->{af} eq AF_INET
-#        ? pack_sockaddr_in(  $port, inet_pton( $self->{af}, $ip ) )
         ? pack_sockaddr_in(  $port, inet_aton( $ip ) )
         : pack_sockaddr_in6( $port, inet_pton( $self->{af}, $ip ) )
     ;
     connect( $connection, $sockaddr ) || $!{EINPROGRESS} || die( format_message( "connect ip = %s, port = %s: %s\n", $ip, $port, $! ) );
 
-    $self->{socket}     = $connection;
-    $self->{_select}    = undef;
-
     # Reset O_NONBLOCK.
     $flags = fcntl( $connection, F_GETFL, 0 ) or die "fcntl F_GETFL: $!\n";  # 0 for error, 0e0 for 0.
     fcntl( $connection, F_SETFL, $flags & ~ O_NONBLOCK ) or die "fcntl F_SETFL not O_NONBLOCK: $!\n";  # 0 for error, 0e0 for 0.
 
     # Use select() to poll for completion or error. When connect succeeds we can write.
-    my $vec = q{};
+    my $vec = '';
     vec( $vec, fileno( $connection ), 1 ) = 1;
     select( undef, $vec, undef, $timeout // $REQUEST_TIMEOUT );
     unless ( vec( $vec, fileno( $connection ), 1 ) ) {
         # If no response yet, impose our own timeout.
-        $! = Errno::ETIMEDOUT();
+        $! = ETIMEDOUT;
         die( format_message( "connect ip = %s, port = %s: %s\n", $ip, $port, $! ) );
     }
 
     # This is how we see whether it connected or there was an error. Document Unix, are you kidding?!
-    $! = unpack( q{L}, getsockopt( $connection, SOL_SOCKET, SO_ERROR ) );
+    $! = unpack( 'L', getsockopt( $connection, SOL_SOCKET, SO_ERROR ) );
     die( format_message( "connect ip = %s, port = %s: %s\n", $ip, $port, $! ) ) if $!;
 
     # Set timeout on all reads and writes.
@@ -487,7 +692,9 @@ sub _connect {
     setsockopt( $connection, SOL_SOCKET, SO_SNDTIMEO, $timeval ) // die "setsockopt SOL_SOCKET, SO_SNDTIMEO: $!\n";
     setsockopt( $connection, SOL_SOCKET, SO_RCVTIMEO, $timeval ) // die "setsockopt SOL_SOCKET, SO_RCVTIMEO: $!\n";
 
-    vec( $self->{_select} = q{}, fileno( $self->{socket} ), 1 ) = 1;
+    $self->{socket} = $connection;
+    my $s = $self->{_io_select} = IO::Select->new;
+    $s->add( $self->{socket} );
 
     return $connection;
 }
@@ -637,7 +844,7 @@ sub _debug_msg {
 
         say STDERR
             "# $header ", $self->{host}, ':', $self->{port}, "\n",
-            '# Hex Stream: ', unpack( q{H*}, $message ), "\n",
+            '# Hex Stream: ', unpack( 'H*', $message ), "\n",
             $_hdr->dump(
                 [
                     [ 'data', length( $message ), $colour ],
@@ -655,10 +862,7 @@ sub _debug_msg {
 # Handler for errors
 sub _error {
     my $self = shift;
-
     Kafka::Exception::IO->throw( throw_args( @_ ) );
-
-    return;
 }
 
 #-- Closes and cleans up -------------------------------------------------------
@@ -687,17 +891,17 @@ Here is the list of possible error messages that C<Kafka::IO> may produce:
 
 Invalid arguments were passed to a method.
 
-=item C<Can't send>
+=item C<Cannot send>
 
-Message can't be sent on a C<Kafka::IO> object socket.
+Message cannot be sent on a C<Kafka::IO> object socket.
 
-=item C<Can't recv>
+=item C<Cannot receive>
 
-Message can't be received.
+Message cannot be received.
 
-=item C<Can't bind>
+=item C<Cannot bind>
 
-TCP connection can't be established on given host and port.
+TCP connection cannot be established on given host and port.
 
 =back
 
@@ -767,6 +971,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
