@@ -6,7 +6,7 @@ Kafka::Producer - Perl interface for Kafka producer client.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::Producer> version 1.001012 .
+This documentation refers to C<Kafka::Producer> version 1.001013 .
 
 =cut
 
@@ -18,7 +18,7 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 #-- load the modules -----------------------------------------------------------
 
@@ -30,10 +30,12 @@ use Params::Util qw(
     _NUMBER
     _STRING
 );
+use Scalar::Util qw(
+    blessed
+);
 use Scalar::Util::Numeric qw(
     isint
 );
-use Try::Tiny;
 
 use Kafka qw(
     %ERROR
@@ -160,19 +162,6 @@ C<new()> takes arguments in key-value pairs. The following arguments are current
 C<$connection> is the L<Kafka::Connection|Kafka::Connection> object responsible for communication with
 the Apache Kafka cluster.
 
-=item C<CorrelationId =E<gt> $correlation_id>
-
-Optional, int32 signed integer, default = C<undef> .
-
-C<Correlation> is a user-supplied integer.
-It will be passed back in the response by the server, unmodified.
-The C<$correlation_id> should be an integer number.
-
-If C<CorrelationId> is not passed to constructor, its value will be assigned automatically
-(random negative integer).
-
-An exception is thrown if C<CorrelationId> sent with request does not match C<CorrelationId> received in response.
-
 =item C<ClientId =E<gt> $client_id>
 
 This is a user supplied identifier (string) for the client application.
@@ -217,7 +206,6 @@ sub new {
 
     my $self = bless {
         Connection      => undef,
-        CorrelationId   => undef,
         ClientId        => undef,
         RequiredAcks    => $WAIT_WRITTEN_TO_LOCAL_LOG,
         Timeout         => $REQUEST_TIMEOUT,
@@ -228,17 +216,14 @@ sub new {
         $self->{ $k } = shift @args if exists $self->{ $k };
     }
 
-    $self->{CorrelationId}  //= _get_CorrelationId;
     $self->{ClientId}       //= 'producer';
 
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'Connection' )
         unless _INSTANCE( $self->{Connection}, 'Kafka::Connection' );
-    $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'CorrelationId (%s)', $self->{CorrelationId} ) )
-        unless isint( $self->{CorrelationId} ) && $self->{CorrelationId} <= $MAX_CORRELATIONID;
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'ClientId' )
         unless ( $self->{ClientId} eq q{} || defined( _STRING( $self->{ClientId} ) ) ) && !utf8::is_utf8( $self->{ClientId} );
     $self->_error( $ERROR_MISMATCH_ARGUMENT, format_message( 'Timeout (%s)', $self->{Timeout} ) )
-        unless _NUMBER( $self->{Timeout} ) && $self->{Timeout} <= $MAX_INT32;
+        unless defined _NUMBER( $self->{Timeout} ) && int( $self->{Timeout} * 1000 ) >= 1 && int( $self->{Timeout} * 1000 ) <= $MAX_INT32;
 
     my $required_acks = $self->{RequiredAcks};
     $self->_error( $ERROR_MISMATCH_ARGUMENT, 'RequiredAcks' )
@@ -338,10 +323,10 @@ sub send {
     my $MessageSet = [];
     my $request = {
         ApiKey                              => $APIKEY_PRODUCE,
-        CorrelationId                       => $self->{CorrelationId},
+        CorrelationId                       => _get_CorrelationId(),
         ClientId                            => $self->{ClientId},
         RequiredAcks                        => $self->{RequiredAcks},
-        Timeout                             => $self->{Timeout} * 1000,
+        Timeout                             => int( $self->{Timeout} * 1000 ),
         topics                              => [
             {
                 TopicName                   => $topic,
@@ -363,7 +348,8 @@ sub send {
         };
     }
 
-    return $self->{Connection}->receive_response_to_request( $request, $compression_codec );
+    my $result = $self->{Connection}->receive_response_to_request( $request, $compression_codec, $self->{Timeout} );
+    return $result;
 }
 
 #-- private attributes ---------------------------------------------------------
@@ -404,19 +390,19 @@ working with L<Kafka|Kafka> package.
 Invalid arguments were provided to a C<new>
 L<constructor|/CONSTRUCTOR> or to other L<method|/METHODS>.
 
-=item C<Can't send>
+=item C<Cannot send>
 
 Request cannot be sent.
 
-=item C<Can't recv>
+=item C<Cannot receive>
 
 Response cannot be received.
 
-=item C<Can't bind>
+=item C<Cannot bind>
 
-TCP connection can't be established on a given host and port.
+TCP connection cannot be established on a given host and port.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 IO error is present, errors found in the structure of the reply or the reply contains a non-zero error codes.
 
@@ -432,7 +418,7 @@ C<CorrelationId> of response doesn't match one in request.
 
 Information about brokers in the cluster is missing.
 
-=item C<Can't get metadata>
+=item C<Cannot get metadata>
 
 Obtained metadata is incorrect or failed to obtain metadata.
 
@@ -481,6 +467,8 @@ L<https://github.com/TrackingSoft/Kafka>
 
 Sergey Gladkov, E<lt>sgladkov@trackingsoft.comE<gt>
 
+Please use GitHub project link above to report problems or contact authors.
+
 =head1 CONTRIBUTORS
 
 Alexander Solovey
