@@ -6,7 +6,7 @@ Kafka::Cluster - object interface to manage a test kafka cluster.
 
 =head1 VERSION
 
-This documentation refers to C<Kafka::Cluster> version 1.001012 .
+This documentation refers to C<Kafka::Cluster> version 1.001013 .
 
 =cut
 
@@ -19,7 +19,7 @@ use warnings;
 
 # ENVIRONMENT ------------------------------------------------------------------
 
-our $VERSION = '1.001012';
+our $VERSION = '1.001013';
 
 use Exporter qw(
     import
@@ -278,7 +278,7 @@ sub new {
     my ( $inifile, $cfg, $zookeeper_client_port, $replication_factor );
 
     opendir( my $dh, $kafka_data_dir )
-        or confess "can't opendir $kafka_data_dir: $!";
+        or confess "cannot opendir $kafka_data_dir: $!";
     foreach my $file ( readdir( $dh ) ) {
         next if $file !~ /^$KAFKA_LOGS_DIR_MASK/;
         ++$replication_factor;
